@@ -7,9 +7,8 @@
 set -Eeuo pipefail
 
 # Configuration:
-# N.B. You may want to also update the index-state in hadrian/cabal.project.
-HACKAGE_INDEX_STATE="2025-01-08T09:21:31Z"
-HAPPY_CONSTRAINT=">=1.20&&<1.21"
+HACKAGE_INDEX_STATE="2020-12-21T14:48:20Z"
+MIN_HAPPY_VERSION="1.20"
 MIN_ALEX_VERSION="3.2.6"
 
 TOP="$(pwd)"
@@ -60,7 +59,7 @@ Environment variables affecting both build systems:
   CROSS_TARGET      Triple of cross-compilation target.
   VERBOSE           Set to non-empty for verbose build output
   RUNTEST_ARGS      Arguments passed to runtest.py
-  MSYSTEM           (Windows-only) Which platform to build form (MINGW64 or MINGW32).
+  MSYSTEM           (Windows-only) Which platform to build from (CLANG64).
   IGNORE_PERF_FAILURES
                     Whether to ignore perf failures (one of "increases",
                     "decreases", or "all")
@@ -142,11 +141,7 @@ function setup_locale() {
 
 function mingw_init() {
   case "$MSYSTEM" in
-    MINGW32)
-      target_triple="i386-unknown-mingw32"
-      boot_triple="i386-unknown-mingw32" # triple of bootstrap GHC
-      ;;
-    MINGW64)
+    CLANG64)
       target_triple="x86_64-unknown-mingw32"
       boot_triple="x86_64-unknown-mingw32" # triple of bootstrap GHC
       ;;
@@ -215,12 +210,10 @@ function set_toolchain_paths() {
           *) fail "unknown NIX_SYSTEM" ;;
         esac
         info "Building toolchain for $NIX_SYSTEM"
-        nix-build .gitlab/darwin/toolchain.nix --argstr system "$NIX_SYSTEM" -o toolchain.sh
+        nix-build --quiet .gitlab/darwin/toolchain.nix --argstr system "$NIX_SYSTEM" -o toolchain.sh
         cat toolchain.sh
       fi
       source toolchain.sh
-      info "--info for GHC for $NIX_SYSTEM"
-      $GHC --info
       ;;
     env)
       # These are generally set by the Docker image but
@@ -279,6 +272,11 @@ function setup() {
   show_tool CABAL
   show_tool HAPPY
   show_tool ALEX
+
+  info "====================================================="
+  info "ghc --info"
+  info "====================================================="
+  $GHC --info
 }
 
 function fetch_ghc() {
@@ -322,8 +320,7 @@ function fetch_cabal() {
         # N.B. Windows uses zip whereas all others use .tar.xz
         MSYS_*|MINGW*)
           case "$MSYSTEM" in
-            MINGW32) cabal_arch="i386" ;;
-            MINGW64) cabal_arch="x86_64" ;;
+            CLANG64) cabal_arch="x86_64" ;;
             *) fail "unknown MSYSTEM $MSYSTEM" ;;
           esac
           url="https://downloads.haskell.org/~cabal/cabal-install-$v/cabal-install-$v-$cabal_arch-windows.zip"
@@ -371,7 +368,7 @@ function setup_toolchain() {
   esac
 
   info "Building happy..."
-  $cabal_install happy --constraint="happy$HAPPY_CONSTRAINT"
+  $cabal_install happy --constraint="happy>=$MIN_HAPPY_VERSION"
 
   info "Building alex..."
   $cabal_install alex --constraint="alex>=$MIN_ALEX_VERSION"
@@ -458,7 +455,7 @@ function push_perf_notes() {
     return
   fi
 
-  if [[ -n "${CROSS_TARGET:-}" ]]; then
+  if [ -n "${CROSS_TARGET:-}" ] && [ "${CROSS_EMULATOR:-}" != "js-emulator" ]; then
     info "Can't test cross-compiled build."
     return
   fi
@@ -517,7 +514,7 @@ function build_hadrian() {
   # hadrian calls tar/xz to produce bindist, there's no other build
   # work taking place.
   if [[ "${CI_JOB_NAME:-}" != *"i386"* ]]; then
-    XZ_OPT="${XZ_OPT:-} -T$cores"
+    export XZ_OPT="${XZ_OPT:-} -T$cores"
   fi
 
   if [[ -n "${REINSTALL_GHC:-}" ]]; then
@@ -609,16 +606,29 @@ function test_hadrian() {
   if [[ "${CROSS_EMULATOR:-}" == "NOT_SET" ]]; then
     info "Cannot test cross-compiled build without CROSS_EMULATOR being set."
     return
+    # special case for JS backend
+  elif [ -n "${CROSS_TARGET:-}" ] && [ "${CROSS_EMULATOR:-}" == "js-emulator" ]; then
+    # The JS backend doesn't support CROSS_EMULATOR logic yet
+    unset CROSS_EMULATOR
+    # run "hadrian test" directly, not using the bindist, even though it did get installed.
+    # This is a temporary solution, See !9515 for the status of hadrian support.
+    run_hadrian \
+      test \
+      --summary-junit=./junit.xml \
+      --test-have-intree-files    \
+      --docs=none                 \
+      "runtest.opts+=${RUNTEST_ARGS:-}" || fail "cross-compiled hadrian main testsuite"
+  elif [[ -n "${CROSS_TARGET:-}" ]] && [[ "${CROSS_TARGET:-}" == *"wasm"* ]]; then
+    run_hadrian \
+      test \
+      --summary-junit=./junit.xml \
+      "runtest.opts+=${RUNTEST_ARGS:-}" || fail "hadrian main testsuite targetting $CROSS_TARGET"
   elif [ -n "${CROSS_TARGET:-}" ]; then
     local instdir="$TOP/_build/install"
     local test_compiler="$instdir/bin/${cross_prefix}ghc$exe"
     install_bindist _build/bindist/ghc-*/ "$instdir"
     echo 'main = putStrLn "hello world"' > expected
     run "$test_compiler" -package ghc "$TOP/.gitlab/hello.hs" -o hello
-    # Despite "-o hello", ghc may output something like hello.exe or
-    # hello.wasm depending on the backend. For the time being let's
-    # just move it to hello before proceeding to running it.
-    mv hello.wasm hello || true
     ${CROSS_EMULATOR:-} ./hello > actual
     run diff expected actual
   elif [[ -n "${REINSTALL_GHC:-}" ]]; then
@@ -801,12 +811,6 @@ function shell() {
   run "$cmd"
 }
 
-# Evaluate some shell code in the CI environment
-function eval_cmd() {
-  local cmd="${@: 1}"
-  eval "$cmd"
-}
-
 function lint_author(){
   base=$1
   head=$2
@@ -936,6 +940,5 @@ case $1 in
   clean) clean ;;
   save_cache) save_cache ;;
   shell) shift; shell "$@" ;;
-  eval) shift; eval_cmd "$@" ;;
   *) fail "unknown mode $1" ;;
 esac
