@@ -26,7 +26,7 @@ import Control.Monad
 import qualified Data.ByteString as BS
 import Data.Foldable
 import Data.Functor
-import qualified GHC.Data.Word64Set as WS
+import qualified Data.IntSet as IS
 import Data.Semigroup
 import Data.String
 import Data.Traversable
@@ -48,6 +48,7 @@ import GHC.Types.ForeignCall
 import GHC.Types.Unique
 import GHC.Types.Unique.FM
 import GHC.Types.Unique.Map
+import GHC.Types.Unique.Supply
 import GHC.Utils.Outputable hiding ((<>))
 import GHC.Utils.Panic
 import GHC.Wasm.ControlFlow.FromCmm
@@ -177,7 +178,7 @@ trouble.
 -}
 globalInfoFromCmmGlobalReg :: WasmTypeTag w -> GlobalReg -> Maybe GlobalInfo
 globalInfoFromCmmGlobalReg t reg = case reg of
-  VanillaReg i _
+  VanillaReg i
     | i >= 1 && i <= 10 -> Just (fromString $ "__R" <> show i, ty_word)
   FloatReg i
     | i >= 1 && i <= 6 ->
@@ -197,7 +198,7 @@ globalInfoFromCmmGlobalReg t reg = case reg of
 
 supportedCmmGlobalRegs :: [GlobalReg]
 supportedCmmGlobalRegs =
-  [VanillaReg i VGcPtr | i <- [1 .. 10]]
+  [VanillaReg i | i <- [1 .. 10]]
     <> [FloatReg i | i <- [1 .. 6]]
     <> [DoubleReg i | i <- [1 .. 6]]
     <> [LongReg i | i <- [1 .. 1]]
@@ -815,7 +816,9 @@ lower_CmmMachOp lbl (MO_SS_Conv w0 w1) xs = lower_MO_SS_Conv lbl w0 w1 xs
 lower_CmmMachOp lbl (MO_UU_Conv w0 w1) xs = lower_MO_UU_Conv lbl w0 w1 xs
 lower_CmmMachOp lbl (MO_XX_Conv w0 w1) xs = lower_MO_UU_Conv lbl w0 w1 xs
 lower_CmmMachOp lbl (MO_FF_Conv w0 w1) xs = lower_MO_FF_Conv lbl w0 w1 xs
-lower_CmmMachOp _ _ _ = panic "lower_CmmMachOp: unreachable"
+lower_CmmMachOp _ mop _ =
+  pprPanic "lower_CmmMachOp: unreachable" $
+    vcat [ text "offending MachOp:" <+> pprMachOp mop ]
 
 -- | Lower a 'CmmLit'. Note that we don't emit 'f32.const' or
 -- 'f64.const' for the time being, and instead emit their relative bit
@@ -872,38 +875,35 @@ lower_CmmReg :: CLabel -> CmmReg -> WasmCodeGenM w (SomeWasmExpr w)
 lower_CmmReg _ (CmmLocal reg) = do
   (reg_i, SomeWasmType ty) <- onCmmLocalReg reg
   pure $ SomeWasmExpr ty $ WasmExpr $ WasmLocalGet ty reg_i
-lower_CmmReg _ (CmmGlobal EagerBlackholeInfo) = do
-  ty_word <- wasmWordTypeM
-  pure $
-    SomeWasmExpr ty_word $
-      WasmExpr $
-        WasmSymConst "__stg_EAGER_BLACKHOLE_info"
-lower_CmmReg _ (CmmGlobal GCEnter1) = do
+lower_CmmReg lbl (CmmGlobal (GlobalRegUse greg reg_use_ty)) = do
   ty_word <- wasmWordTypeM
   ty_word_cmm <- wasmWordCmmTypeM
-  onFuncSym "__stg_gc_enter_1" [] [ty_word_cmm]
-  pure $ SomeWasmExpr ty_word $ WasmExpr $ WasmSymConst "__stg_gc_enter_1"
-lower_CmmReg _ (CmmGlobal GCFun) = do
-  ty_word <- wasmWordTypeM
-  ty_word_cmm <- wasmWordCmmTypeM
-  onFuncSym "__stg_gc_fun" [] [ty_word_cmm]
-  pure $ SomeWasmExpr ty_word $ WasmExpr $ WasmSymConst "__stg_gc_fun"
-lower_CmmReg lbl (CmmGlobal BaseReg) = do
-  platform <- wasmPlatformM
-  lower_CmmExpr lbl $ regTableOffset platform 0
-lower_CmmReg lbl (CmmGlobal reg) = do
-  ty_word <- wasmWordTypeM
-  if
+  case greg of
+    EagerBlackholeInfo ->
+      pure $
+        SomeWasmExpr ty_word $
+          WasmExpr $
+            WasmSymConst "__stg_EAGER_BLACKHOLE_info"
+    GCEnter1 -> do
+      onFuncSym "__stg_gc_enter_1" [] [ty_word_cmm]
+      pure $ SomeWasmExpr ty_word $ WasmExpr $ WasmSymConst "__stg_gc_enter_1"
+    GCFun -> do
+      onFuncSym "__stg_gc_fun" [] [ty_word_cmm]
+      pure $ SomeWasmExpr ty_word $ WasmExpr $ WasmSymConst "__stg_gc_fun"
+    BaseReg -> do
+      platform <- wasmPlatformM
+      lower_CmmExpr lbl $ regTableOffset platform 0
+    _other
       | Just (sym_global, SomeWasmType ty) <-
-          globalInfoFromCmmGlobalReg ty_word reg ->
+          globalInfoFromCmmGlobalReg ty_word greg ->
           pure $ SomeWasmExpr ty $ WasmExpr $ WasmGlobalGet ty sym_global
       | otherwise -> do
           platform <- wasmPlatformM
-          case someWasmTypeFromCmmType $ globalRegType platform reg of
+          case someWasmTypeFromCmmType reg_use_ty of
             SomeWasmType ty -> do
               (WasmExpr ptr_instr, o) <-
                 lower_CmmExpr_Ptr lbl $
-                  get_GlobalReg_addr platform reg
+                  get_GlobalReg_addr platform greg
               pure $
                 SomeWasmExpr ty $
                   WasmExpr $
@@ -1331,7 +1331,7 @@ lower_CmmUnsafeForeignCall_Drop ::
   [CmmActual] ->
   WasmCodeGenM w (WasmStatements w)
 lower_CmmUnsafeForeignCall_Drop lbl sym_callee ret_cmm_ty arg_exprs = do
-  ret_uniq <- wasmUniq
+  ret_uniq <- getUniqueM
   let ret_local = LocalReg ret_uniq ret_cmm_ty
   lower_CmmUnsafeForeignCall
     lbl
@@ -1379,7 +1379,7 @@ lower_CmmUnsafeForeignCall lbl target mb_hints ret_info ret_locals arg_exprs = d
           (reg_i, SomeWasmType reg_ty) <- onCmmLocalReg reg
           pure $
             SomeWasmPostCCall (reg_ty `TypeListCons` acc_tys) $
-              case (# ret_hint, cmmRegWidth platform $ CmmLocal reg #) of
+              case (# ret_hint, cmmRegWidth $ CmmLocal reg #) of
                 (# SignedHint, W8 #) ->
                   acc_instr
                     `WasmConcat` WasmConst reg_ty 0xFF
@@ -1458,7 +1458,7 @@ lower_CmmAction lbl act = do
       (i, SomeWasmType ty_reg) <- onCmmLocalReg reg
       WasmExpr instrs <- lower_CmmExpr_Typed lbl ty_reg e
       pure $ WasmStatements $ instrs `WasmConcat` WasmLocalSet ty_reg i
-    CmmAssign (CmmGlobal reg) e
+    CmmAssign (CmmGlobal (GlobalRegUse reg _)) e
       | BaseReg <- reg -> pure $ WasmStatements WasmNop
       | Just (sym_global, SomeWasmType ty_reg) <-
           globalInfoFromCmmGlobalReg ty_word reg -> do
@@ -1531,9 +1531,11 @@ lower_CmmGraph :: CLabel -> CmmGraph -> WasmCodeGenM w (FuncBody w)
 lower_CmmGraph lbl g = do
   ty_word <- wasmWordTypeM
   platform <- wasmPlatformM
+  us <- getUniqueSupplyM
   body <-
     structuredControl
       platform
+      us
       (\_ -> lower_CmmExpr_Typed lbl ty_word)
       (lower_CmmActions lbl)
       g
@@ -1551,7 +1553,7 @@ onTopSym lbl = case sym_vis of
   SymDefault -> wasmModifyM $ \s ->
     s
       { defaultSyms =
-          WS.insert
+          IS.insert
             (getKey $ getUnique sym)
             $ defaultSyms s
       }
