@@ -17,7 +17,7 @@ module GHC.Driver.CmdLine
       Flag(..), defFlag, defGhcFlag, defGhciFlag, defHiddenFlag, hoistFlag,
       errorsToGhcException,
 
-      Err(..), Warn(..), WarnReason(..),
+      Err(..), Warn, warnsToMessages,
 
       EwM, runEwM, addErr, addWarn, addFlagWarn, getArg, getCurLoc, liftEwM
     ) where
@@ -25,18 +25,18 @@ module GHC.Driver.CmdLine
 import GHC.Prelude
 
 import GHC.Utils.Misc
-import GHC.Utils.Outputable
 import GHC.Utils.Panic
 import GHC.Utils.Panic.Plain
 import GHC.Data.Bag
 import GHC.Types.SrcLoc
-import GHC.Utils.Json
-
-import GHC.Types.Error ( DiagnosticReason(..) )
+import GHC.Types.Error
+import GHC.Utils.Error
+import GHC.Driver.Errors.Types
+import GHC.Driver.Errors.Ppr ()
+import GHC.Utils.Outputable (text)
 
 import Data.Function
 import Data.List (sortBy, intercalate, stripPrefix)
-import Data.Word
 
 import GHC.ResponseFile
 import Control.Exception (IOException, catch)
@@ -75,7 +75,7 @@ hoistFlag f (Flag a b c) = Flag a (go b) c
       go (OptPrefix k) = OptPrefix (\s -> go2 (k s))
       go (OptIntSuffix k) = OptIntSuffix (\n -> go2 (k n))
       go (IntSuffix k) = IntSuffix (\n -> go2 (k n))
-      go (Word64Suffix k) = Word64Suffix (\s -> go2 (k s))
+      go (WordSuffix k) = WordSuffix (\s -> go2 (k s))
       go (FloatSuffix k) = FloatSuffix (\s -> go2 (k s))
       go (PassFlag k) = PassFlag (\s -> go2 (k s))
       go (AnySuffix k) = AnySuffix (\s -> go2 (k s))
@@ -98,7 +98,7 @@ data OptKind m                             -- Suppose the flag is -f
     | OptPrefix (String -> EwM m ())       -- -f or -farg (i.e. the arg is optional)
     | OptIntSuffix (Maybe Int -> EwM m ()) -- -f or -f=n; pass n to fn
     | IntSuffix (Int -> EwM m ())          -- -f or -f=n; pass n to fn
-    | Word64Suffix (Word64 -> EwM m ())    -- -f or -f=n; pass n to fn
+    | WordSuffix (Word -> EwM m ())        -- -f or -f=n; pass n to fn
     | FloatSuffix (Float -> EwM m ())      -- -f or -f=n; pass n to fn
     | PassFlag  (String -> EwM m ())       -- -f; pass "-f" fn
     | AnySuffix (String -> EwM m ())       -- -f or -farg; pass entire "-farg" to fn
@@ -108,32 +108,16 @@ data OptKind m                             -- Suppose the flag is -f
 --         The EwM monad
 --------------------------------------------------------
 
--- | Used when filtering warnings: if a reason is given
--- it can be filtered out when displaying.
-data WarnReason
-  = NoReason
-  | ReasonDeprecatedFlag
-  | ReasonUnrecognisedFlag
-  deriving (Eq, Show)
-
-instance Outputable WarnReason where
-  ppr = text . show
-
-instance ToJson WarnReason where
-  json NoReason = JSNull
-  json reason   = JSString $ show reason
-
 -- | A command-line error message
 newtype Err  = Err { errMsg :: Located String }
 
 -- | A command-line warning message and the reason it arose
-data Warn = Warn
-  {   warnReason :: DiagnosticReason,
-      warnMsg    :: Located String
-  }
+--
+-- This used to be own type, but now it's just @'MsgEnvelope' 'DriverMessage'@.
+type Warn = Located DriverMessage
 
 type Errs  = Bag Err
-type Warns = Bag Warn
+type Warns = [Warn]
 
 -- EwM ("errors and warnings monad") is a monad
 -- transformer for m that adds an (err, warn) state
@@ -153,7 +137,7 @@ instance MonadIO m => MonadIO (EwM m) where
     liftIO = liftEwM . liftIO
 
 runEwM :: EwM m a -> m (Errs, Warns, a)
-runEwM action = unEwM action (panic "processArgs: no arg yet") emptyBag emptyBag
+runEwM action = unEwM action (panic "processArgs: no arg yet") emptyBag mempty
 
 setArg :: Located String -> EwM m () -> EwM m ()
 setArg l (EwM f) = EwM (\_ es ws -> f l es ws)
@@ -162,11 +146,12 @@ addErr :: Monad m => String -> EwM m ()
 addErr e = EwM (\(L loc _) es ws -> return (es `snocBag` Err (L loc e), ws, ()))
 
 addWarn :: Monad m => String -> EwM m ()
-addWarn = addFlagWarn WarningWithoutFlag
+addWarn msg = addFlagWarn $ DriverUnknownMessage $ mkSimpleUnknownDiagnostic $
+  mkPlainDiagnostic WarningWithoutFlag noHints $ text msg
 
-addFlagWarn :: Monad m => DiagnosticReason -> String -> EwM m ()
-addFlagWarn reason msg = EwM $
-  (\(L loc _) es ws -> return (es, ws `snocBag` Warn reason (L loc msg), ()))
+addFlagWarn :: Monad m => DriverMessage -> EwM m ()
+addFlagWarn msg = EwM
+  (\(L loc _) es ws -> return (es, L loc msg : ws, ()))
 
 getArg :: Monad m => EwM m String
 getArg = EwM (\(L _ arg) es ws -> return (es, ws, arg))
@@ -177,6 +162,10 @@ getCurLoc = EwM (\(L loc _) es ws -> return (es, ws, loc))
 liftEwM :: Monad m => m a -> EwM m a
 liftEwM action = EwM (\_ es ws -> do { r <- action; return (es, ws, r) })
 
+warnsToMessages :: DiagOpts -> [Warn] -> Messages DriverMessage
+warnsToMessages diag_opts = foldr
+  (\(L loc w) ws -> addMessage (mkPlainMsgEnvelope diag_opts loc w) ws)
+  emptyMessages
 
 --------------------------------------------------------
 --         Processing arguments
@@ -188,10 +177,10 @@ processArgs :: Monad m
             -> (FilePath -> EwM m [Located String]) -- ^ response file handler
             -> m ( [Located String],  -- spare args
                    [Err],  -- errors
-                   [Warn] ) -- warnings
+                   Warns ) -- warnings
 processArgs spec args handleRespFile = do
     (errs, warns, spare) <- runEwM action
-    return (spare, bagToList errs, bagToList warns)
+    return (spare, bagToList errs, warns)
   where
     action = process args []
 
@@ -251,7 +240,7 @@ processOneArg opt_kind rest arg args
         IntSuffix f | Just n <- parseInt rest_no_eq -> Right (f n, args)
                     | otherwise -> Left ("malformed integer argument in " ++ dash_arg)
 
-        Word64Suffix f | Just n <- parseWord64 rest_no_eq -> Right (f n, args)
+        WordSuffix f | Just n <- parseWord rest_no_eq -> Right (f n, args)
                      | otherwise -> Left ("malformed natural argument in " ++ dash_arg)
 
         FloatSuffix f | Just n <- parseFloat rest_no_eq -> Right (f n, args)
@@ -280,7 +269,7 @@ arg_ok (Prefix          _)  _    _   = True -- Missing argument checked for in p
                                             -- to improve error message (#12625)
 arg_ok (OptIntSuffix    _)  _    _   = True
 arg_ok (IntSuffix       _)  _    _   = True
-arg_ok (Word64Suffix    _)  _    _   = True
+arg_ok (WordSuffix      _)  _    _   = True
 arg_ok (FloatSuffix     _)  _    _   = True
 arg_ok (OptPrefix       _)  _    _   = True
 arg_ok (PassFlag        _)  rest _   = null rest
@@ -296,8 +285,8 @@ parseInt s = case reads s of
                  ((n,""):_) -> Just n
                  _          -> Nothing
 
-parseWord64 :: String -> Maybe Word64
-parseWord64 s = case reads s of
+parseWord :: String -> Maybe Word
+parseWord s = case reads s of
                  ((n,""):_) -> Just n
                  _          -> Nothing
 
