@@ -117,7 +117,7 @@ import {-# SOURCE #-}   GHC.Types.Id.Info( IdDetails, IdInfo, coVarDetails, isCo
 import {-# SOURCE #-}   GHC.Builtin.Types ( manyDataConTy )
 import GHC.Types.Name hiding (varName)
 import GHC.Types.Unique ( Uniquable, Unique, getKey, getUnique
-                        , nonDetCmpUnique )
+                        , mkUniqueGrimily, nonDetCmpUnique )
 import GHC.Types.Basic( TypeOrConstraint(..) )
 import GHC.Utils.Misc
 import GHC.Utils.Binary
@@ -247,7 +247,7 @@ data Var
   = TyVar {  -- Type and kind variables
              -- see Note [Kind and type variables]
         varName    :: !Name,
-        realUnique :: {-# UNPACK #-} !Unique,
+        realUnique :: {-# UNPACK #-} !Int,
                                      -- ^ Key for fast comparison
                                      -- Identical to the Unique in the name,
                                      -- cached here for speed
@@ -258,14 +258,14 @@ data Var
                                         -- Used for kind variables during
                                         -- inference, as well
         varName        :: !Name,
-        realUnique     :: {-# UNPACK #-} !Unique,
+        realUnique     :: {-# UNPACK #-} !Int,
         varType        :: Kind,
         tc_tv_details  :: TcTyVarDetails
   }
 
   | Id {
         varName    :: !Name,
-        realUnique :: {-# UNPACK #-} !Unique,
+        realUnique :: {-# UNPACK #-} !Int,
         varType    :: Type,
         varMult    :: Mult,             -- See Note [Multiplicity of let binders]
         idScope    :: IdScope,
@@ -374,10 +374,10 @@ instance Eq Var where
     a == b = realUnique a == realUnique b
 
 instance Ord Var where
-    a <= b = getKey (realUnique a) <= getKey (realUnique b)
-    a <  b = getKey (realUnique a) <  getKey (realUnique b)
-    a >= b = getKey (realUnique a) >= getKey (realUnique b)
-    a >  b = getKey (realUnique a) >  getKey (realUnique b)
+    a <= b = realUnique a <= realUnique b
+    a <  b = realUnique a <  realUnique b
+    a >= b = realUnique a >= realUnique b
+    a >  b = realUnique a >  realUnique b
     a `compare` b = a `nonDetCmpVar` b
 
 -- | Compare Vars by their Uniques.
@@ -397,7 +397,7 @@ instance HasOccName Var where
   occName = nameOccName . varName
 
 varUnique :: Var -> Unique
-varUnique var = realUnique var
+varUnique var = mkUniqueGrimily (realUnique var)
 
 varMultMaybe :: Id -> Maybe Mult
 varMultMaybe (Id { varMult = mult }) = Just mult
@@ -405,12 +405,12 @@ varMultMaybe _ = Nothing
 
 setVarUnique :: Var -> Unique -> Var
 setVarUnique var uniq
-  = var { realUnique = uniq,
+  = var { realUnique = getKey uniq,
           varName = setNameUnique (varName var) uniq }
 
 setVarName :: Var -> Name -> Var
 setVarName var new_name
-  = var { realUnique = getUnique new_name,
+  = var { realUnique = getKey (getUnique new_name),
           varName = new_name }
 
 setVarType :: Var -> Type -> Var
@@ -687,10 +687,6 @@ Currently there are nine different uses of 'VarBndr':
 * TyCon.TyConBinder = VarBndr TyVar TyConBndrVis
   Binders of a TyCon; see TyCon in GHC.Core.TyCon
 
-* TyCon.TyConPiTyBinder = VarBndr TyCoVar TyConBndrVis
-  Binders of a PromotedDataCon
-  See Note [Promoted GADT data constructors] in GHC.Core.TyCon
-
 * IfaceType.IfaceForAllBndr     = VarBndr IfaceBndr ForAllTyFlag
 * IfaceType.IfaceForAllSpecBndr = VarBndr IfaceBndr Specificity
 * IfaceType.IfaceTyConBinder    = VarBndr IfaceBndr TyConBndrVis
@@ -1084,7 +1080,7 @@ updateTyVarKindM update tv
 
 mkTyVar :: Name -> Kind -> TyVar
 mkTyVar name kind = TyVar { varName    = name
-                          , realUnique = nameUnique name
+                          , realUnique = getKey (nameUnique name)
                           , varType  = kind
                           }
 
@@ -1092,7 +1088,7 @@ mkTcTyVar :: Name -> Kind -> TcTyVarDetails -> TyVar
 mkTcTyVar name kind details
   = -- NB: 'kind' may be a coercion kind; cf, 'GHC.Tc.Utils.TcMType.newMetaCoVar'
     TcTyVar {   varName    = name,
-                realUnique = nameUnique name,
+                realUnique = getKey (nameUnique name),
                 varType  = kind,
                 tc_tv_details = details
         }
@@ -1148,7 +1144,7 @@ mkExportedLocalVar details name ty info
 mk_id :: Name -> Mult -> Type -> IdScope -> IdDetails -> IdInfo -> Id
 mk_id name !w ty scope details info
   = Id { varName    = name,
-         realUnique = nameUnique name,
+         realUnique = getKey (nameUnique name),
          varMult    = w,
          varType    = ty,
          idScope    = scope,
