@@ -13,7 +13,7 @@ HC=${3:-ghc}
 mode=${4:-norm}
 
 EXE=""
-if [[ $(shell uname -s | grep -c 'MSYS\|MINGW') -eq 1 ]]
+if [[ $(uname -s | grep -c 'MSYS\|MINGW') -eq 1 ]]
 then
   EXE=".exe"
 fi
@@ -49,4 +49,4 @@ ${output}/fasta-c${EXE} $MODE_ARGS | tr -d '\r' > ${output}/${INPUT_FILE}
 # 	./fasta-c $(NORM_OPTS) | tr -d '\r' > $@
 
 # k-nucleotide.slowstdin : fasta-c
-# 	./fasta-c $(SLOW_OPTS) | tr -d '\r' > $@
\ No newline at end of file
+# 	./fasta-c $(SLOW_OPTS) | tr -d '\r' > $@
