@@ -117,7 +117,8 @@ void postSparkCountersEvent (Capability *cap,
  */
 void postThreadLabel(Capability    *cap,
                      EventThreadID  id,
-                     char          *label);
+                     char          *label,
+                     size_t         len);
 
 /*
  * Various GC and heap events
@@ -189,17 +190,11 @@ void postProfSampleCostCentre(Capability *cap,
 void postProfBegin(void);
 #endif /* PROFILING */
 
-void postIPE(StgWord64 info,
-             const char *table_name,
-             const char *closure_desc,
-             const char *ty_desc,
-             const char *label,
-             const char *module,
-             const char *srcloc);
+void postIPE(const InfoProvEnt *ipe);
 
 void postConcUpdRemSetFlush(Capability *cap);
 void postConcMarkEnd(StgWord32 marked_obj_count);
-void postNonmovingHeapCensus(int log_blk_size,
+void postNonmovingHeapCensus(uint16_t blk_size,
                              const struct NonmovingAllocCensus *census);
 
 #if defined(TICKY_TICKY)
